"""
Tests for OpenFermion to pyQuil object Translator
"""
import pytest
import numpy as np
from openfermion.ops import (FermionOperator, hermitian_conjugated,
                             QubitOperator, InteractionOperator,
                             InteractionRDM)
from openfermion.transforms import jordan_wigner
from forestopenfermion.pyquil_connector import (pyquilpauli_to_qubitop,
                                                qubitop_to_pyquilpauli)
from pyquil.paulis import PauliTerm, PauliSum


def test_confirm_interface():
    """
    Test conversion

    This checks the conversion between pyQuil PauliTerms/Sums
    with OpenFermion QubitOperators. This checks for fields associated with the
    conversion between the two objects. Does not check the field types.
    """
    op = QubitOperator()
    assert hasattr(op, 'terms')
    assert isinstance(op.terms, dict)

    # PauliSum does not allow initialization with an empty value
    op_pyquil = PauliSum([PauliTerm("X", 1)])
    assert hasattr(op_pyquil, 'terms')
    # yell if Forest paulis data structure changes
    assert hasattr(PauliTerm("X", 1), '_ops')


def test_translation_type_enforcement():
    """
    Make sure type check works
    """
    create_one = FermionOperator('1^')
    empty_one_body = np.zeros((2, 2))
    empty_two_body = np.zeros((2, 2, 2, 2))
    interact_one = InteractionOperator(1, empty_one_body, empty_two_body)
    interact_rdm = InteractionRDM(empty_one_body, empty_two_body)

    with pytest.raises(TypeError):
        qubitop_to_pyquilpauli(create_one)
    with pytest.raises(TypeError):
        qubitop_to_pyquilpauli(interact_one)
    with pytest.raises(TypeError):
        qubitop_to_pyquilpauli(interact_rdm)

    # don't accept anything other than pyquil PauliSum or PauliTerm
    with pytest.raises(TypeError):
        qubitop_to_pyquilpauli(create_one)
    with pytest.raises(TypeError):
        qubitop_to_pyquilpauli(interact_one)
    with pytest.raises(TypeError):
        qubitop_to_pyquilpauli(interact_rdm)


def test_qubitop_to_paulisum():
    """
    Conversion of QubitOperator; accuracy test
    """
    hop_term = FermionOperator(((2, 1), (0, 0)))
    term = hop_term + hermitian_conjugated(hop_term)

    pauli_term = jordan_wigner(term)

    forest_term = qubitop_to_pyquilpauli(pauli_term)
    ground_truth = PauliTerm("X", 0)*PauliTerm("Z", 1)*PauliTerm("X", 2)
    ground_truth += PauliTerm("Y", 0)*PauliTerm("Z", 1)*PauliTerm("Y", 2)
    ground_truth *= 0.5

    assert ground_truth == forest_term


def test_qubitop_to_paulisum_zero():

    identity_term = QubitOperator()
    forest_term = qubitop_to_pyquilpauli(identity_term)
    ground_truth = PauliTerm("I", 0, 0)

    assert ground_truth == forest_term


def test_pyquil_to_qubitop():
    pyquil_term = PauliSum([PauliTerm("X", 0)*PauliTerm("Y", 5)])
    op_fermion_term = QubitOperator(((0, "X"), (5, "Y")))
    test_op_fermion_term = pyquilpauli_to_qubitop(pyquil_term)
    assert test_op_fermion_term.isclose(op_fermion_term)


def test_pyquil_to_qubitop_type_enforced():
    """Enforce the appropriate type"""
    create_one = FermionOperator('1^')
    empty_one_body = np.zeros((2, 2))
    empty_two_body = np.zeros((2, 2, 2, 2))
    interact_one = InteractionOperator(1, empty_one_body, empty_two_body)
    interact_rdm = InteractionRDM(empty_one_body, empty_two_body)

    with pytest.raises(TypeError):
        pyquilpauli_to_qubitop(create_one)
    with pytest.raises(TypeError):
        pyquilpauli_to_qubitop(interact_one)
    with pytest.raises(TypeError):
        pyquilpauli_to_qubitop(interact_rdm)

    # don't accept anything other than pyquil PauliSum or PauliTerm
    with pytest.raises(TypeError):
        pyquilpauli_to_qubitop(create_one)
    with pytest.raises(TypeError):
        pyquilpauli_to_qubitop(interact_one)
    with pytest.raises(TypeError):
        pyquilpauli_to_qubitop(interact_rdm)


def test_pyquil_to_qubitop_pauliterm_conversion():
    """Test if a pauliterm is converted to a pauli sum"""
    pyquil_term = PauliTerm("X", 0)*PauliTerm("Y", 5)
    # implicit test of conversion from Term to Sum
    open_fermion_term = pyquilpauli_to_qubitop(pyquil_term)
    op_fermion_term = QubitOperator(((0, "X"), (5, "Y")))
    assert open_fermion_term.isclose(op_fermion_term)



############################################################################
#   Copyright 2017 Rigetti Computing, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
############################################################################
"""
An interface from OpenFermion QubitObjects to some of the circuit generating functionality in pyquil
"""
from forestopenfermion.pyquil_connector import qubitop_to_pyquilpauli
from openfermion.ops import QubitOperator

from pyquil.quil import Program
from pyquil.paulis import exponentiate as pyquil_exponentiate


def exponentiate(qubit_operator):
    """
    Generates a pyquil program corresponding to the QubitOperator generator.

    The OpenFermion qubit operator is translated to a pyQuil PauliSum which, in turn, is passed to
    the `exponentiate' method. The `exponentiate' method generates a circuit that can be simulated
    with the Forest-qvm or associated QVMs.

    :param QubitOperator qubit_operator: Generator of rotations
    :return: a pyQuil program representing the unitary evolution
    :rtype: Program
    """
    if not isinstance(qubit_operator, QubitOperator):
        raise TypeError("qubit_operator must be an OpenFermion "
                        "QubitOperator type")

    pauli_sum_representation = qubitop_to_pyquilpauli(qubit_operator)
    prog = Program()
    for term in pauli_sum_representation.terms:
        prog += pyquil_exponentiate(term)

    return prog


def TimeEvolution(time, hamiltonian):
    """
    Time evolve a hamiltonian

    Converts the Hamiltonian to an instance of the pyQuil Pauliterms and returns the time evolution
    operator. This method mirrors the ProjectQ TimeEvolution interface.

    :param [float, int] time: time to evolve
    :param QubitOperator hamiltonian: a Hamiltonian as a OpenFermion QubitOperator
    :return: a pyquil Program representing the Hamiltonian
    :rtype: Program
    """
    if not isinstance(time, (int, float)):
        raise TypeError("float must be a float or an int")
    if not isinstance(hamiltonian, QubitOperator):
        raise TypeError("hamiltonian must be an OpenFermion "
                        "QubitOperator object")

    pyquil_pauli_term = qubitop_to_pyquilpauli(hamiltonian)
    pyquil_pauli_term *= time
    prog = Program()
    for term in pyquil_pauli_term.terms:
        prog += pyquil_exponentiate(term)

    return prog



############################################################################
#   Copyright 2017 Rigetti Computing, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
############################################################################
from setuptools import setup, find_packages
from forestopenfermion import __version__

# Readme file as long_description:
long_description = open('README.rst').read()

setup(
    name='forestopenfermion',
    version=__version__,
    author='Rigetti Computing',
    author_email='softapps@rigetti.com',
    description='A plugin allowing OpenFermion to interface with Forest.',
    long_description=long_description,
    license='Apache 2',
    install_requires=[
        'scipy >= 0.18.0',
        'numpy >= 1.11.0',
        'openfermion == 0.1a0',
        'pyquil >= 1.1.0',
        'future'
    ],
    packages=find_packages(exclude=["tests"])
)
